#include <stdint.h>
#include <string.h>

#include "C3D/sprites.h"
#include "C3D/fixed.h"
#include "C3D/game.h"

#include "game.h"
#include "gfx.h"
#include "map.h"

int sreserve_c = 0;
Sprite sprite_reserve[SINDEX_S];

Rect sprite_colls[] = {
  {},
  {},
  {},
  {},
  {.x0 = fix(.25), .y0 = fix(.25), .x1 = fix(.5), .y1 = fix(.5)},
  {},
  {},
  {}
};

int16_t sprite_hp[] = {
  0x1000,
  0x1000,
  0x1000,
  0x1000,
  2,
  0x1000,
  0x1000,
  0x1000
};

Sprite *cc_mksprite(V2d pos, int type){
  if(sreserve_c >= SINDEX_S)
    return NULL;
  Sprite *spr = &sprite_reserve[sreserve_c++];
  spr->tex = type;
  spr->coll = sprite_colls[type];
  spr->pos = pos;
  spr->hp = sprite_hp[type];
  add_sprite(spr);
  return spr;
}

int cc_rmsprite(int pos){
  if(pos < 0 || pos >= sreserve_c)
    return 1;

  if(!remove_sprite(&sprite_reserve[pos]))
    return 2;
  if(pos+1 != sreserve_c){
    memmove(&sprite_reserve[pos], &sprite_reserve[pos+1], sizeof(Sprite)*(sreserve_c-pos));
  } 
  sreserve_c--;
  return 0;
}

void cc_load_msprites(Sprite *mapsprites, int spritec){
  sreserve_c = 0;
  for(int i = 0; i < spritec; i++)
    cc_mksprite(mapsprites[i].pos, mapsprites[i].tex);
}

void cc_hit(CCGame *game, int hitstr){
  if(game->hit > 0)
    return;
  game->hit = 40;
  game->_bgame.player.hp -= hitstr;  
}

void cc_ai_mov(RcGame *game, Sprite *spr, fixed_t spd, int hitstr, int coll, fixed_t range){
  fixed_t dx = spr->pos.x - game->player.pos.x;
  fixed_t dy = spr->pos.y - game->player.pos.y;
  fixed_t player_dist = fsq(dx) + fsq(dy);
  int dxs = (dx >= 0) * 2 - 1;
  int dys = (dy >= 0) * 2 - 1; 
    
  if(player_dist > fsq(max_dist) || player_dist < fsq(fix(0.5)))
    return;
  if(player_dist < fsq(range))
    cc_hit((CCGame*)game, hitstr);
  
  int nx = spr->pos.x;
  int ny = spr->pos.y;
  if(fixabs(dx) >= fixabs(dy))
    nx -= dxs * spd;
  else
    ny -= dys * spd;
  if(!game->current_map->dat[ffloor(nx)*game->current_map->w + ffloor(ny)] || !coll){  
    spr->pos.x = nx;
    spr->pos.y = ny;
  }

  return;
}

int cc_ai(RcGame *game){
  for(int i = 0; i < sreserve_c; i++){
    Sprite *spr = &sprite_reserve[i];
    if(spr->hp <= 0){
      if(cc_rmsprite(i) == 2)
        game->flags.exit = 2;
      continue;
    }
    switch(spr->tex){
      default:
      case 0:
      case 1:
      case 2:
      case 3:
        break;
      case 4:{
        cc_ai_mov(game, spr, fix(.025), 5, 1, fix(1));
        break;
      }      
    }
  }
  return 0;
}

WeaponStat weapon_stats[] = {
  {0},
  {.range = fix(1), .dmg = 1},
  {.range = fix(1.5), .dmg = 5}
};

void weapon_hurt_stuff(CCGame *game){
  Sprite *hit;
  int dist = raycast((RcGame*)game, &((RcGame*)game)->player, &hit);
  if(dist > weapon_stats[game->curr_weapon].range || !hit)
    return;
  //:p
  if(hit > 0x90000000)
    return;
  hit->hp -= weapon_stats[game->curr_weapon].dmg;
}

extern WeapAnim weap_anims[WEAP_ANIM_C]; 

int cc_logic(RcGame *_game){
  CCGame *game = (void*)_game;
  if(game->hit)
    game->hit--;
  
  if(game->weapon_use){
    game->weapon_use = (game->weapon_use+1) % 
                         (weap_anims[game->curr_weapon].framec*4);
    if(game->weapon_use == weap_anims[game->curr_weapon].frame_hit*4)
      weapon_hurt_stuff(game);
  }

  if(_game->player.hp <= 0){
    death_anim(_game);
    switch_map(game, game->curr_map);
  }
  
  return 0;
}

void use_weapon(CCGame *game){
  if(game->weapon_use)
    return;
  game->weapon_use = 1;
}

void init_logic(){
  add_logic_hook((RcLogicFunc*)&cc_ai);
  add_logic_hook((RcLogicFunc*)&cc_logic);
}
