#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/clock.h>
#include <gint/image.h>
#include <gint/gray.h>
#include <gint/dma.h>
#include <gint/gdb.h>
#include <libprof.h>

#include "C3D/fixed.h"
#include "C3D/game.h"
#include "C3D/sprites.h"
#include "C3D/moteur.h"
#include "C3D/map.h"
#include "C3D/config.h"

#include "gfx.h"
#include "map.h"
#include "game.h"

//====== Copy3DEngine =====
// Git du moteur : https://git.planet-casio.com/Fcalva/Copy3DEngine
// Git du jeu : [Rajoutez le vôtre ici]
//
// Page du jeu : [La vôtre ici]
//
// Voir README.md pour license précise, par Fcalva et est sous GPLv3
// 
//

extern bopti_image_t *tex_index[TINDEX_S];

char disp_frame_time = 0;
char first_frame = 0;
int frame_time_timer = 1;

CCGame game = {
  ._bgame = {
    .player = {
      .pos = {fix(3.1), fix(3.1)},
      .dir = {fix(0), fix(1)},
      .plane = {fix(0.66), fix(0)},
      .hp = 100
    },
    .flags = {0, 0}
  },
  .hit = 0,
  .curr_weapon = 1,
  .weapon_use = 0
};

int frame_time = 1;

void keys_get(RcGame *game){
  pollevent();
  
  move(game);

  if(keydown(KEY_SHIFT)){
    use_weapon((CCGame*)game); 
  }

  if (keydown(KEY_F1) && frame_time_timer <= 0) {
    if (disp_frame_time == 0) {
      disp_frame_time = 1;
      frame_time_timer = 10;
    }
    else {
      disp_frame_time = 0;
      frame_time_timer = 10;
    }
  }
  frame_time_timer--;
  if (keydown(KEY_EXIT)) game->flags.exit = 1;

  if(keydown(KEY_F2)){
    dprint(0,0,3,"%d",raycast(game, &game->player, NULL));
  }

  if(keydown(KEY_F3)){
   cc_mksprite(game->player.pos, 7); 
   while(keydown(KEY_F3)){
     clearevents();
     pollevent();
   }
  }
  if(keydown(KEY_F4)){
    cc_rmsprite(0);
    while(keydown(KEY_F4)) { clearevents(); pollevent();}
  }

  #ifdef debug
  //if (keydown(KEY_TAN)) end_screen();
  #endif
}

extern bopti_image_t title;

int main(){
  RcGame *rcgame = (RcGame*)&game;

  dclear(C_WHITE);

  //trucs de chargement

  load_map();

  prof_init();

  if(dgray(DGRAY_ON))
    rcgame->flags.exit = 1;
  dupdate();

  dimage(0,0,&title);
  dupdate();
  sleep_us(2000000);

  #if debug
  EngineTimers timers;
  //gdb_start_on_exception();
  #endif

  init_gfx();
  init_map(&game);
  init_logic(); 
 
  while (!rcgame->flags.exit) {
    prof_t frame = prof_make();
    prof_enter(frame);

    dclear(C_WHITE);

    keys_get(rcgame);
   
    do_logic(rcgame, frame_time);

    draw_walls(
    #if debug
      &timers,
    #endif
      rcgame
    );

    draw_sprites(tex_index, &rcgame->player);

    if (disp_frame_time == 1) dprint( 0, 0, C_BLACK, "%d ms", frame_time/1000);

    draw_gui(&game);

    #if debug
    dprint( 1, 27, C_BLACK, "pX %d", ffloor(rcgame->player.pos.x));
    dprint( 1, 36, C_BLACK, "pY %d", ffloor(rcgame->player.pos.y));
    dprint( 1, 9, C_BLACK, "Rct %d", prof_time(timers.raycast_time));
    dprint( 1, 18, C_BLACK, "Dt %d", prof_time(timers.draw_time));
    timers.raycast_time = prof_make();
    timers.draw_time = prof_make();
    #endif
    if(game.hit > 30){
      drect(0,0,127,63,C_INVERT);
    }
    dupdate();
    prof_leave(frame);
    frame_time = prof_time(frame);
    first_frame = 0;
  }

  dgray(DGRAY_OFF);

  prof_quit();

  if(rcgame->flags.exit > 1){
    dclear(0);
    dprint(0,0,3,"Error %x !", rcgame->flags.exit);
    dupdate();
    getkey(); 
  }
  return 1;
}
